unit DoublePoints;

interface

uses Windows, SysUtils, Classes;

type
  PLongintPoint = ^TLongintPoint;
  TLongintPoint = record
    X, Y: Longint;
  end;
  TLongintRect = record
    case Integer of
      0: (Left, Top, Right, Bottom: Longint);
      1: (TopLeft, BottomRight: TLongintPoint);
  end;
  PBooleanPoint = ^TBooleanPoint;
  TBooleanPoint = record
    X, Y: Boolean;
  end;
  PDoublePoint = ^TDoublePoint;
  TDoublePoint = record
    X, Y: Double;
  end;
  PDoubleRect = ^TDoubleRect;
  TDoubleRect = record
    case Integer of
      0: (Left, Top, Right, Bottom: Double);
      1: (TopLeft, BottomRight: TDoublePoint);
  end;
  TQuadrilateral = array [0..3] of TDoublePoint;
  PQuadrilateral = ^TQuadrilateral;
  PStringPoint = ^TStringPoint;
  TStringPoint = record
    X, Y: string;
  end;

function LPCreate(const X, Y: Longint): TLongintPoint;
function LPDistance(const A, B: TLongintPoint): Double;
function LPAdd(const A, B: TLongintPoint): TLongintPoint;
function LPSubtract(const A, B: TLongintPoint): TLongintPoint;
function LPMultiply(const A: TLongintPoint; const B: Longint): TLongintPoint;
function LPNormalize(const A: TLongintPoint): TLongintPoint;
function LPMag(const A: TLongintPoint): Double;
function LPToWP(const P: TLongintPoint): TPoint;
function BPCreate(const X, Y: Boolean): TBooleanPoint;
function DPCreate(const X, Y: Double): TDoublePoint;
function DPAdd(const A, B: TDoublePoint): TDoublePoint;
function DPSubtract(const A, B: TDoublePoint): TDoublePoint;
function DPMultiply(const A: TDoublePoint; const B: Double): TDoublePoint;
function DPDivide(const A: TDoublePoint; const B: Double): TDoublePoint;
function DPCross(const A, B: TDoublePoint): Double;
function DPDot(const A, B: TDoublePoint): Double;
function DPMag(const A: TDoublePoint): Double;
function DPTheta(const A: TDoublePoint): Double;
function DPRotate(const A: TDoublePoint; const Theta: Double): TDoublePoint;
function DPNormal(const A: TDoublePoint): TDoublePoint;
function DPRound(const A: TDoublePoint): TLongintPoint;
function DPDistance(const A, B: TDoublePoint): Double;
function DPNormalizeMax(const A: TDoublePoint): TDoublePoint;
function DPNormalize(const A: TDoublePoint): TDoublePoint;
function DPPixelize(const A: TDoublePoint): TDoublePoint;
function SPCreate(const X, Y: string): TStringPoint;
function IsPointInRect(const P: TDoublePoint; const Rect: TDoubleRect): Boolean;
function IsPointInQuadrilateral(const P: TDoublePoint; const Q: TQuadrilateral): Boolean;

const
  DPUndefined: TDoublePoint = (X: 0.0; Y: 0.0);

type
  TComparisonResult = (crLessThan, crEqualTo, crGreaterThan);
  TComparisonRoutine = function(const A, B: TDoublePoint): TComparisonResult;

  TDoublePointArray = class(TPersistent)
  protected
    FPoints: Pointer;
    FCount: Longint;
    FSize: Longint;
    FAllocSize: Longint;
    procedure SetSize(const NewSize: Longint);
    procedure CheckBounds(const Index: Longint);
    function GetPoint(const Index: Longint): TDoublePoint;
    function PointerTo(const Index: Longint): PDoublePoint;
    procedure Pack;
  public
    constructor Create(const AllocAmount: Longint);
    destructor Destroy; override;
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    procedure Assign(Source: TPersistent); override;
    function First: TDoublePoint;
    function Last: TDoublePoint;
    procedure Append(const ADoublePoint: TDoublePoint);
    procedure Remove(const Index: Longint);
    procedure RemoveRange(const StartIndex, EndIndex: Longint);
    procedure AppendRange(const Source: TDoublePointArray; const StartIndex, EndIndex: Longint);
    procedure RemoveByIndex(SortedIndeces: TList);
    procedure Insert(const P: TDoublePoint; Index: Longint);
    procedure InsertPoints(const Source: TDoublePointArray; Index: Longint);
    procedure Resize(const NewCount: Longint);
    procedure Clear;
    procedure SortXAscending;
    procedure SortYAscending;
    procedure SortXDescending;
    procedure SortYDescending;
    procedure Sort(const Routine: TComparisonRoutine);
    function IndexOf(const P: TDoublePoint; const Tolerance: Double): Integer;
    property Count: Longint read FCount;
    property Points[const Index: Longint]: TDoublePoint read GetPoint; default;
    property RawData: Pointer read FPoints;
  end;

implementation

uses Math;

function LPCreate(const X, Y: Longint): TLongintPoint;
begin
  Result.X := X;
  Result.Y := Y;
end;

function LPDistance(const A, B: TLongintPoint): Double;
begin
  Result := Hypot(A.X - B.X, A.Y - B.Y);
end;

function LPAdd(const A, B: TLongintPoint): TLongintPoint;
begin
  Result.X := A.X + B.X;
  Result.Y := A.Y + B.Y;
end;

function LPSubtract(const A, B: TLongintPoint): TLongintPoint;
begin
  Result.X := A.X - B.X;
  Result.Y := A.Y - B.Y;
end;

function LPMultiply(const A: TLongintPoint; const B: Longint): TLongintPoint;
begin
  Result.X := A.X * B;
  Result.Y := A.Y * B;
end;

function LPNormalize(const A: TLongintPoint): TLongintPoint;
var
  H: Double;
begin
  if (A.X = 0) and (A.Y = 0) then begin
    // Convention only.
    Result.X := 1;
    Result.Y := 0;
  end
  else begin
    H := LPMag(A);
    if H > 1e-15 then begin
      Result.X := Round(A.X / H);
      Result.Y := Round(A.Y / H);
    end;
    if (Result.X = 0) and (Result.Y = 0) then begin
      // Convention only.
      Result.X := 1;
      Result.Y := 0;
    end;
  end;
end;

function LPMag(const A: TLongintPoint): Double;
begin
  Result := Hypot(A.X, A.Y);
end;

function LPToWP(const P: TLongintPoint): TPoint;
begin
  Result.X := P.X;
  Result.Y := P.Y;
end;

function BPCreate(const X, Y: Boolean): TBooleanPoint;
begin
  Result.X := X;
  Result.Y := Y;
end;

function DPCreate(const X, Y: Double): TDoublePoint;
begin
  Result.X := X;
  Result.Y := Y;
end;

function DPAdd(const A, B: TDoublePoint): TDoublePoint;
begin
  Result.X := A.X + B.X;
  Result.Y := A.Y + B.Y;
end;

function DPSubtract(const A, B: TDoublePoint): TDoublePoint;
begin
  Result.X := A.X - B.X;
  Result.Y := A.Y - B.Y;
end;

function DPMultiply(const A: TDoublePoint; const B: Double): TDoublePoint;
begin
  Result.X := A.X * B;
  Result.Y := A.Y * B;
end;

function DPDivide(const A: TDoublePoint; const B: Double): TDoublePoint;
begin
  Result.X := A.X / B;
  Result.Y := A.Y / B;
end;

function DPCross(const A, B: TDoublePoint): Double;
begin
  Result := A.X * B.Y - A.Y * B.X;
end;

function DPDot(const A, B: TDoublePoint): Double;
begin
  Result := A.X * B.X + A.Y * B.Y;
end;

function DPMag(const A: TDoublePoint): Double;
begin
  Result := Hypot(A.X, A.Y);
end;

function DPTheta(const A: TDoublePoint): Double;
{
const
  InfiniteTangent: array [Boolean] of Double = ( -Pi / 2, Pi / 2 );
}
begin
  {
  if A.X = 0 then
    Result := InfiniteTangent[A.Y >= 0]
  else begin
    Result := ArcTan(A.Y / A.X);
    if A.X < 0 then Result := Result + Pi;
  }
    Result := ArcTan2(A.Y, A.X);
  {
  end;
  }
end;

function DPRotate(const A: TDoublePoint; const Theta: Double): TDoublePoint;
var
  T, CosTheta, SinTheta: Double;
begin
  CosTheta := Cos(Theta);
  SinTheta := Sin(Theta);
  T := A.X;
  Result.X := A.X * CosTheta + A.Y * SinTheta;
  Result.Y := A.Y * CosTheta -   T * SinTheta;
end;

function DPNormal(const A: TDoublePoint): TDoublePoint;
begin
  Result.Y := A.X;
  Result.X := -A.Y;
end;

function DPRound(const A: TDoublePoint): TLongintPoint;
begin
  Result.X := Round(A.X);
  Result.Y := Round(A.Y);
end;

function DPDistance(const A, B: TDoublePoint): Double;
begin
  Result := Hypot(A.X - B.X, A.Y - B.Y);
end;

function DPNormalizeMax(const A: TDoublePoint): TDoublePoint;
var
  M: Double;
begin
  M := Math.Max(Abs(A.X), Abs(A.Y));
  Result.X := A.X / M;
  Result.Y := A.Y / M;
end;

function DPNormalize(const A: TDoublePoint): TDoublePoint;
var
  H: Double;
begin
  Result := A;
  H := DPMag(Result);
  if H > 1e-15 then begin
    Result.X := A.X / H;
    Result.Y := A.Y / H;
  end;
end;

function DPPixelize(const A: TDoublePoint): TDoublePoint;
begin
  Result := DPNormalize(A);
  Result.X := Result.X * Sqrt(2);
  Result.Y := Result.Y * Sqrt(2);
end;

constructor TDoublePointArray.Create(const AllocAmount: Longint);
begin
  inherited Create;
  FCount := 0;
  FSize := 0;
  FPoints := nil;
  FAllocSize := AllocAmount * SizeOf(TDoublePoint);
end;

destructor TDoublePointArray.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TDoublePointArray.LoadFromStream(Stream: TStream);
begin
  Clear;
  Stream.Read(FCount, SizeOf(FCount));
  Resize(FCount);
  if FCount > 0 then Stream.Read(FPoints^, FSize);
end;

procedure TDoublePointArray.SaveToStream(Stream: TStream);
begin
  Stream.Write(FCount, SizeOf(FCount));
  if FCount > 0 then Stream.Write(FPoints^, FCount * SizeOf(TDoublePoint));
end;

procedure TDoublePointArray.Assign(Source: TPersistent);
var
  DPA: TDoublePointArray;
begin
  if Source is TDoublePointArray then begin
    DPA := Source as TDoublePointArray;
    FAllocSize := DPA.FAllocSize;
    FCount := DPA.FCount;
    SetSize(DPA.FSize);
    Move(DPA.FPoints^, FPoints^, FSize);
  end
  else begin
    inherited;
  end;
end;

procedure TDoublePointArray.SetSize(const NewSize: Longint);
begin
  if FSize <> NewSize then begin
    ReallocMem(FPoints, NewSize);
    FSize := NewSize;
  end;
end;

procedure TDoublePointArray.CheckBounds(const Index: Longint);
begin
  if (Index < 0) or (Index > FCount - 1) then begin
    raise Exception.Create('TDoublePointArray: Index out of bounds.');
  end;
end;

function TDoublePointArray.GetPoint(const Index: Longint): TDoublePoint;
var
  P: PDoublePoint;
begin
  CheckBounds(Index);
  P := FPoints;
  Inc(P, Index);
  Result := P^;
end;

function TDoublePointArray.PointerTo(const Index: Longint): PDoublePoint;
begin
  CheckBounds(Index);
  Result := FPoints;
  Inc(Result, Index);
end;

procedure TDoublePointArray.Append(const ADoublePoint: TDoublePoint);
var
  P: PDoublePoint;
begin
  if FCount * SizeOf(TDoublePoint) = FSize then SetSize(FSize + FAllocSize);
  P := FPoints;
  Inc(P, FCount);
  P^ := ADoublePoint;
  Inc(FCount);
end;

procedure TDoublePointArray.Clear;
begin
  FCount := 0;
  SetSize(0);
end;

procedure TDoublePointArray.Resize(const NewCount: Longint);
begin
  SetSize(NewCount * SizeOf(TDoublePoint));
end;

procedure TDoublePointArray.Remove(const Index: Longint);
begin
  RemoveRange(Index, Index);
end;

procedure TDoublePointArray.RemoveRange(const StartIndex, EndIndex: Longint);
var
  StartP, EndP: PDoublePoint;
begin
  CheckBounds(StartIndex);
  CheckBounds(EndIndex);
  if EndIndex < StartIndex then Exit;
  StartP := FPoints;
  Inc(StartP, StartIndex);
  EndP := FPoints;
  Inc(EndP, EndIndex + 1);
  Move(EndP^, StartP^, (FCount - EndIndex) * SizeOf(TDoublePoint));
  Dec(FCount, EndIndex - StartIndex + 1);
end;

procedure TDoublePointArray.AppendRange(const Source: TDoublePointArray;
                                        const StartIndex, EndIndex: Longint);
var
  SourceP, DestP: PDoublePoint;
  PointCount: Integer;
begin
  Source.CheckBounds(StartIndex);
  Source.CheckBounds(EndIndex);
  if EndIndex < StartIndex then Exit;
  SourceP := Source.FPoints;
  Inc(SourceP, StartIndex);
  PointCount := EndIndex - StartIndex + 1;
  Resize(Count + PointCount);
  DestP := FPoints;
  Inc(DestP, Count);
  Move(SourceP^, DestP^, PointCount * SizeOf(TDoublePoint));
  Inc(FCount, PointCount);
end;

procedure TDoublePointArray.Pack;
begin
end;

procedure TDoublePointArray.RemoveByIndex(SortedIndeces: TList);
var
  I, Index: Longint;
begin
  for I := 0 to SortedIndeces.Count - 1 do begin
    Index := Longint(SortedIndeces[I]);
    Remove(Index - I);
  end;
end;

procedure TDoublePointArray.Insert(const P: TDoublePoint; Index: Longint);
var
  StartP, EndP: PDoublePoint;
begin
  Resize(Count + 1);
  StartP := FPoints;
  Inc(StartP, Index);
  EndP := StartP;
  Inc(EndP, 1);
  Move(StartP^, EndP^, (Count - Index) * SizeOf(TDoublePoint));
  Move(P, StartP^, SizeOf(P));
  Inc(FCount, 1);
end;

procedure TDoublePointArray.InsertPoints(const Source: TDoublePointArray; Index: Longint);
var
  StartP, EndP: PDoublePoint;
begin
  Resize(Count + Source.Count);
  StartP := FPoints;
  Inc(StartP, Index);
  EndP := StartP;
  Inc(EndP, Source.Count);
  Move(StartP^, EndP^, (Count - Index) * SizeOf(TDoublePoint));
  Move(Source.FPoints^, StartP^, Source.Count * SizeOf(TDoublePoint));
  Inc(FCount, Source.Count);
end;

{------------------------- TSortableDoublePointArray --------------------------}

function CompareXValuesAscending(const A, B: TDoublePoint): TComparisonResult;
begin
  if A.X > B.X then
    Result := crGreaterThan
  else if A.X < B.X then
    Result := crLessThan
  else begin
    Result := crEqualTo;
  end;
end;

function CompareXValuesDescending(const A, B: TDoublePoint): TComparisonResult;
begin
   if A.X > B.X then
      Result := crLessThan
   else if A.X < B.X then
      Result := crGreaterThan
   else begin
      Result := crEqualTo;
   end;
end;

function CompareYValuesAscending(const A, B: TDoublePoint): TComparisonResult;
begin
  if A.Y > B.Y then
    Result := crGreaterThan
  else if A.Y < B.Y then
    Result := crLessThan
  else begin
    Result := crEqualTo;
  end;
end;

function CompareYValuesDescending(const A, B: TDoublePoint): TComparisonResult;
begin
   if A.Y > B.Y then
      Result := crLessThan
   else if A.Y < B.Y then
      Result := crGreaterThan
   else begin
      Result := crEqualTo;
   end;
end;

procedure TDoublePointArray.SortXAscending;
begin
  Sort(CompareXValuesAscending);
end;

procedure TDoublePointArray.SortXDescending;
begin
  Sort(CompareXValuesDescending);
end;

procedure TDoublePointArray.SortYAscending;
begin
  Sort(CompareYValuesAscending);
end;

procedure TDoublePointArray.SortYDescending;
begin
  Sort(CompareYValuesDescending);
end;

procedure TDoublePointArray.Sort(const Routine: TComparisonRoutine);
var
  I, IR, J, L: Longint;
  RRA: TDoublePoint;
begin
  if Count < 2 then Exit;
  L := Count shr 1 + 1;
  IR := Count;
  // The index l will be decremented from its initial value down to 1 during
  // the "hiring" phase. Once it reaches 1, the index ir will be decremented
  // from its initial value down to 1 during the "retirement-and-promotion"
  // phase.
  while True do begin
    if L > 1 then begin
      Dec(L);
      RRA := Points[L - 1];
    end
    else begin
      RRA := Points[IR - 1];
      PointerTo(IR - 1)^ := Points[0];
      Dec(IR);
      if IR = 1 then begin
        PointerTo(0)^ := RRA;
        Break;
      end;
    end;
    I := L;
    J := L + L;
    while J <= IR do begin
      if (J < IR) and (Routine(Points[J - 1], Points[J]) = crLessThan) then Inc(J);
      if Routine(RRA, Points[J - 1]) = crLessThan then begin
        PointerTo(I - 1)^ := Points[J - 1];
        I := J;
        J := J shl 1;
      end
      else begin
        Break;
      end;
    end;
    PointerTo(I - 1)^ := RRA;
  end;
end;

function TDoublePointArray.IndexOf(const P: TDoublePoint;
                                   const Tolerance: Double): Integer;
var
  PP: PDoublePoint;
begin
  Result := Count - 1;
  PP := FPoints;
  Inc(PP, Result);
  while Result > -1 do begin
    if DPDistance(PP^, P) < Tolerance then Break;
    Dec(Result);
    Dec(PP);
  end;
end;

function TDoublePointArray.First: TDoublePoint;
begin
  Result := Self[0];
end;

function TDoublePointArray.Last: TDoublePoint;
begin
  Result := Self[Count - 1];
end;

function SPCreate(const X, Y: string): TStringPoint;
begin
  Result.X := X;
  Result.Y := Y;
end;

function CorrectedRect(const Rect: TDoubleRect): TDoubleRect;
begin
  if Rect.Left > Rect.Right then begin
    Result.Left := Rect.Right;
    Result.Right := Rect.Left;
  end
  else begin
    Result.Left := Rect.Left;
    Result.Right := Rect.Right;
  end;
  if Rect.Bottom > Rect.Top then begin
    Result.Top := Rect.Bottom;
    Result.Bottom := Rect.Top;
  end
  else begin
    Result.Top := Rect.Top;
    Result.Bottom := Rect.Bottom;
  end;
end;

function IsPointInRect(const P: TDoublePoint; const Rect: TDoubleRect): Boolean;
var
  CR: TDoubleRect;
begin
  CR := CorrectedRect(Rect);
  Result := (CR.Left <= P.X) and (P.X <= CR.Right) and
            (CR.Bottom <= P.Y) and (P.Y <= CR.Top);
end;

function IsPointInQuadrilateral(const P: TDoublePoint; const Q: TQuadrilateral): Boolean;
var
  I, J: Integer;
begin
  Result := False;
  J := High(Q);
  for I := Low(Q) to High(Q) do begin
    if ((
         ((Q[I].Y <= P.Y) and (P.Y < Q[J].Y)) or
         ((Q[J].Y <= P.Y) and (P.Y < Q[I].Y))
        ) and
        (P.X < (Q[J].X - Q[I].X) * (P.Y - Q[I].Y) / (Q[J].Y - Q[I].Y) + Q[I].X)) then
    begin
      Result := not Result;
    end;
    J := I;
  end;
end;

end.
